# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from abc import ABC, abstractmethod
from typing import Any, Callable, final

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.distributed import ReduceOp

from libinephany.observations import observation_utils
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.utils import torch_distributed_utils
from libinephany.utils.constants import PARAMS, SCHEDULER_GROUP_NAME
from libinephany.utils.torch_distributed_utils import MASTER_SCHEDULER_RANK
from libinephany.utils.torch_utils import ACTIVATIONS

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

StatisticsStorage = list[torch.Tensor] | list[TensorStatistics] | list[float]

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Statistic(ABC):

    def __init__(
        self,
        *,
        parameter_group_name: str,
        averaging_function: Callable[[list[float], float, str], float],
        agent_modules: dict[str, str],
        max_statistic_cache_size: int = 3,
        tensor_stats_downsample_percentage: float = 0.01,
        statistic_sample_frequency: int = 10,
        invalid_observation_threshold: float = 1e10,
        **kwargs,
    ) -> None:
        """
        :param parameter_group: Name of the parameter group to gather statistics from.
        :param averaging_function: Averaging function to apply over the metrics.
        :param agent_modules: Dictionary mapping agent controlled module names to the type of module being controlled.
        :param max_statistic_cache_size: Maximum number of tensors to store internally before processing the tensors
        into statistics.
        :param tensor_stats_downsample_percentage: Percentage of elements of the tensors from the model to randomly
        sample to compute the statistics from.
        :param statistic_sample_frequency: How frequently to cache tensors from the model for later statistic
        computation.
        :param invalid_observation_threshold: Threshold for invalid observations, default is 1e10.
        :param kwargs: Other observation keyword arguments.
        """

        self._tensor_cache: list[torch.Tensor] = []
        self._data: StatisticsStorage = []

        self._averaging_function = averaging_function
        self._sample_number = 0

        self.parameter_group_name = parameter_group_name
        self.agent_modules = agent_modules
        self.max_cache_size = max_statistic_cache_size
        self.downsample_percent = tensor_stats_downsample_percentage
        self.sample_frequency = statistic_sample_frequency
        self.include_statistics: list[str] | None = None
        self.invalid_observation_threshold = invalid_observation_threshold

    @final
    @property
    def tracker_name(self) -> str:
        """
        :return: Name of the tracker.
        """

        return f"{self.__class__.__name__}-{self.parameter_group_name}"

    @final
    @property
    def storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        storage_format = self._get_storage_format()

        if storage_format in StatisticStorageTypes:
            return storage_format

        raise ValueError(
            f"The type of storage format statistic trackers should return must be in the "
            f"{StatisticStorageTypes.__name__} enumeration class. {storage_format} is not and so "
            f"{self.__class__.__name__} is invalid!"
        )

    @property
    def uses_forward_hook(self) -> bool:
        """
        :return: Whether this local observation relies on PyTorch's register_forward_hook.
        """

        return False

    @property
    def requires_gradient_graphs(self) -> bool:
        """
        :return: Whether the tracker requires gradient graphs to be retained.
        """

        return False

    @property
    def hook_module_type_names(self) -> list[str]:
        """
        :return: Names of the module types to hook the observation into.
        """

        raise NotImplementedError

    def _get_forward_hook(self) -> Callable[[nn.Module, torch.Tensor, torch.Tensor], None]:
        """
        :return: Forward hook to register the function with.
        """

        raise NotImplementedError

    @abstractmethod
    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :return: None, TensorStatistics model or a float.
        """

        raise NotImplementedError

    @abstractmethod
    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        raise NotImplementedError

    @staticmethod
    def _get_parameters(parameter_group: dict[str, Any]) -> list[torch.Tensor]:
        """
        :param parameter_group: Parameter group to extract parameter tensors and name from.
        :return: List of the parameter tensors for the given parameter group.
        """

        return [p for p in parameter_group[PARAMS] if p.requires_grad]

    def _find_parameter_group(self, optimizer: optim.Optimizer) -> dict[str, Any]:
        """
        :param optimizer: Optimizer storing references to all parameter groups in the (possibly) distributed model.
        :return: Parameter group corresponding to the parameter group named by the parameter_group_name instance
        variable.
        :raises KeyError: If a parameter group with the stored parameter_group_name could not be found in the optimizer.
        """

        for param_group in optimizer.param_groups:
            if SCHEDULER_GROUP_NAME in param_group:
                group_name = param_group[SCHEDULER_GROUP_NAME]

                if group_name == self.parameter_group_name:
                    return param_group

        raise KeyError(f"Could not find parameter group with name {self.parameter_group_name}.")

    def _process_tensor_cache(self) -> None:
        """
        Processes the tensor cache to build a TensorStatistic model.
        """

        if not self.include_statistics:
            raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")

        if self._tensor_cache:
            concatenated = torch.cat(self._tensor_cache)
            self._tensor_cache = []

            statistics = TensorStatistics.build(
                tensor=concatenated,
                include_statistics=self.include_statistics,
                sample_percentage=self.downsample_percent,
            )
            self._data.append(statistics)  # type: ignore

    @staticmethod
    @final
    def _determine_reduction_shape(statistic: torch.Tensor | TensorStatistics | float | None) -> list[int] | None:
        """
        :param statistic: Statistic to get the shape for as defined by the master rank.
        :return: Shape of the statistic according to the master rank.
        """

        shape = None

        if torch_distributed_utils.is_scheduler_master_rank():
            if isinstance(statistic, torch.Tensor):
                shape = statistic.view(-1).shape

            elif isinstance(statistic, TensorStatistics):
                shape = statistic.to_tensor().view(-1).shape

            elif statistic is not None:
                shape = torch.tensor([statistic]).shape

        broadcasted_shape = list(shape) if shape is not None else None
        broadcasted_shape = torch_distributed_utils.broadcast_data(data=broadcasted_shape)

        return broadcasted_shape

    @final
    def _distributed_reduce(
        self, statistic: torch.Tensor | TensorStatistics | float | None
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param statistic: Statistic to reduce across all distributed ranks. This can be a Torch tensor, a float or a
        TensorStatistics object.
        :return: None if this rank is not the master rank or the reduced statistic if it is.
        """

        if not torch_distributed_utils.is_distributed():
            return statistic

        shape = self._determine_reduction_shape(statistic=statistic)

        if statistic is None:
            to_reduce = torch.zeros(shape, dtype=torch.float64)

        elif isinstance(statistic, torch.Tensor):
            to_reduce = statistic.clone().to(torch.float64).view(-1)

        elif isinstance(statistic, TensorStatistics):
            to_reduce = statistic.to_tensor().to(torch.float64).view(-1)

        else:
            to_reduce = torch.tensor([statistic], dtype=torch.float64)

        to_reduce = to_reduce.to(torch_distributed_utils.get_local_device())
        dist.reduce(to_reduce, dst=MASTER_SCHEDULER_RANK, op=ReduceOp.SUM)

        if not torch_distributed_utils.is_scheduler_master_rank():
            return None

        if isinstance(statistic, TensorStatistics):
            return TensorStatistics.from_tensor(tensor=to_reduce)

        elif statistic is not None and not isinstance(statistic, torch.Tensor):
            return to_reduce.item()

        return to_reduce

    @final
    def gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
    ) -> None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        """

        parameter_group = self._find_parameter_group(optimizer=optimizer)
        parameters = self._get_parameters(parameter_group=parameter_group)

        if self._sample_number % self.sample_frequency == 0:
            statistic = self._gather(
                optimizer=optimizer, model=model, parameters=parameters, parameter_group=parameter_group
            )

            statistic = self._distributed_reduce(statistic=statistic)

            if torch_distributed_utils.is_scheduler_master_rank():
                if isinstance(statistic, torch.Tensor):
                    statistic = statistic.view(-1)
                    self._tensor_cache.append(statistic)

                    if len(self._tensor_cache) >= self.max_cache_size:
                        self._process_tensor_cache()

                elif statistic is not None:
                    self._data.append(statistic)  # type: ignore

        self._sample_number += 1

    @final
    def fetch(self) -> TensorStatistics | float | None:
        """
        :return: None if this method is called on a rank other than the master rank otherwise a dictionary mapping
        parameter group names to TensorStatistic models or floats.
        """

        if not torch_distributed_utils.is_scheduler_master_rank():
            return None

        if self.storage_format is StatisticStorageTypes.TENSOR_STATISTICS:
            self._process_tensor_cache()

            results = observation_utils.apply_averaging_function_to_tensor_statistics(
                tensor_statistics=self._data,  # type: ignore
                averaging_function=self._averaging_function,
                invalid_value_threshold=self.invalid_observation_threshold,
                tracker_name=self.tracker_name,
            )

        elif self.storage_format is StatisticStorageTypes.FLOAT:
            results = self._averaging_function(
                self._data,  # type: ignore
                invalid_value_threshold=self.invalid_observation_threshold,
                tracker_name=self.tracker_name,
            )

        else:
            raise ValueError(f"Data storage type {self.storage_format} is invalid!")

        self.reset()

        return results  # type: ignore

    @final
    def register(self, model: nn.Module) -> None:
        """
        :param model: Model to register the hooks with.
        """

        if not self.uses_forward_hook:
            raise RuntimeError(
                f"The LocalObservation {self.__class__.__name__} does not use forward hooks. The register method should"
                f"not have been called!"
            )

        for name, module in model.named_modules():
            if self.parameter_group_name not in name:
                continue

            if module.__class__.__name__ not in self.hook_module_type_names:
                continue

            module.register_forward_hook(self._get_forward_hook())

    def reset(self) -> None:
        """
        Clears any internal states.
        """

        self._tensor_cache = []
        self._data = []
        self._sample_number = 0


class FirstOrderGradients(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.include_statistics = include_statistics

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        gradients = [
            p.grad.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        if not gradients:
            return None

        stacked_grads = torch.cat(gradients)

        return stacked_grads


class SecondOrderGradients(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        compute_hessian_diagonal: bool = False,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param compute_hessian_diagonal: Whether to compute the Hessian diagonal to determine second order gradients
        or use the squared first order gradients as approximations in the same way Adam does.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.compute_hessian_diagonal = compute_hessian_diagonal
        self.include_statistics = include_statistics

    @property
    def requires_gradient_graphs(self) -> bool:
        """
        :return: Whether the statistic requires gradient graphs to be retained.
        """

        return self.compute_hessian_diagonal

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    @staticmethod
    def compute_hessian_diagonals(parameters: list[torch.Tensor]) -> torch.Tensor:
        """
        :param parameters: Parameters to compute the hessian diagonal matrices for.
        :return: Tensor containing the hessian diagonal matrices for all given parameters.
        """

        hessian_diagonals = []

        for parameter in parameters:
            if parameter.grad is not None:
                so_gradient = torch.autograd.grad(
                    outputs=parameter.grad.clone(),
                    inputs=parameter,
                    grad_outputs=torch.ones_like(parameter.grad, requires_grad=True),
                    only_inputs=True,
                    retain_graph=True,
                    create_graph=True,
                    allow_unused=True,
                )[0]

                if so_gradient is not None:
                    hessian_diagonals.append(so_gradient.view(-1))
                else:
                    hessian_diagonals.append(torch.zeros_like(parameter.view(-1)))

        return torch.cat(hessian_diagonals)

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        fo_gradients = [
            p.grad.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        if not fo_gradients:
            return None

        if self.compute_hessian_diagonal:
            so_grad_tensor = self.compute_hessian_diagonals(parameters=parameters)

        else:
            so_grad_tensor = torch.cat(fo_gradients) ** 2

        return so_grad_tensor


class ActivationStatistics(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.include_statistics = include_statistics

    @property
    def uses_forward_hook(self) -> bool:
        """
        :return: Whether this local observation relies on PyTorch's register_forward_hook.
        """

        return True

    @property
    def hook_module_type_names(self) -> list[str]:
        """
        :return: Names of the module types to hook the observation into.
        """

        return ACTIVATIONS

    def _get_forward_hook(self) -> Callable[[nn.Module, torch.Tensor, torch.Tensor], None]:
        """
        :return: Forward hook to register the function with.
        """

        if self.include_statistics is None:
            raise ValueError("include_statistics is required to use forward hooks!")

        def hook(module: nn.Module, layer_input: torch.Tensor, layer_output: torch.Tensor) -> None:
            """
            :param module: Module the hook was registered with. Not used here.
            :param layer_input: Input to the module. Not used here.
            :param layer_output: Output of the module.
            """

            if self._sample_number % self.sample_frequency == 0:
                statistics = TensorStatistics.build(
                    tensor=layer_output,
                    include_statistics=self.include_statistics,
                    sample_percentage=self.downsample_percent,
                )
                self._data.append(statistics)  # type: ignore

            self._sample_number += 1

        return hook

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None.
        """

        return None


class ParameterUpdateStatistics(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.include_statistics = include_statistics

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        update_tensor = observation_utils.form_update_tensor(
            optimizer=optimizer, parameters=parameters, parameter_group=parameter_group
        )

        if update_tensor is None:
            update_tensor = torch.cat([torch.zeros(p.view(-1).shape, device=p.device) for p in parameters])

        return update_tensor


class LHOPTParameterUpdateStatistics(ParameterUpdateStatistics):

    pass


class ParameterStatistics(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.include_statistics = include_statistics

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        return torch.cat([p.data.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p)])


class LHOPTParameterStatistics(ParameterStatistics):

    pass


class LAMBTrustRatioStatistics(Statistic):

    def __init__(
        self,
        *,
        include_statistics: list[str] | None = None,
        use_log_transform: bool = False,
        **kwargs,
    ) -> None:
        """
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param use_log_transform: Whether to transform the LAMB trust ratio by taking ln(1 + R).
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.include_statistics = include_statistics
        self.use_log_transform = use_log_transform

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        weights_list = [p.data.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p)]
        if weights_list:
            weights = torch.cat(weights_list)

        else:
            weights = None

        updates = observation_utils.form_update_tensor(
            optimizer=optimizer, parameters=parameters, parameter_group=parameter_group
        )

        update_norm = torch.norm(updates, p=2).item() if updates is not None else 0
        weight_norm = torch.norm(weights, p=2).item() if weights is not None else 0

        lamb_trust_ratio = 0.0
        if update_norm > 0:
            lamb_trust_ratio = weight_norm / update_norm

            if self.use_log_transform:
                lamb_trust_ratio = math.log(1 + lamb_trust_ratio)

        return lamb_trust_ratio


class LHOPTLAMBTrustRatioStatistics(LAMBTrustRatioStatistics):

    pass


class NumberOfParameters(Statistic):

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self._count: int | None = None

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        if not torch_distributed_utils.is_scheduler_master_rank():
            return None

        if self._count is not None:
            return self._count

        count = sum(parameter.numel() for parameter in model.parameters())
        self._count = count

        return count


class NumberOfLayers(Statistic):

    def __init__(
        self,
        *,
        trainable_only: bool = False,
        **kwargs,
    ) -> None:
        """
        :param use_log_transform: Whether to transform the number of layers by ln(1 + N).
        :param trainable_only: Whether to only count trainable layers.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self._count: int | None = None

        self.trainable_only = trainable_only

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def count_layers_recursively(self, module: nn.Module):
        """
        :param module: The PyTorch model or module to count the number of layers of.
        :return: Total count of layers matching the criteria.
        """

        count = 0

        for submodule in module.children():
            if not list(submodule.children()):
                if self.trainable_only:
                    if any(p.requires_grad for p in submodule.parameters()):
                        count += 1
                else:
                    count += 1

            else:
                count += self.count_layers_recursively(submodule)

        return count

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        if not torch_distributed_utils.is_scheduler_master_rank():
            return None

        if self._count is not None:
            return self._count

        count = self.count_layers_recursively(module=model)
        self._count = count

        return count


class GradientVarianceFraction(Statistic):

    def __init__(
        self,
        *,
        variance_threshold: float = LHOPT_CONSTANTS["DEFAULT_VARIANCE_THRESHOLD"],
        **kwargs,
    ) -> None:
        """
        :param variance_threshold: Threshold for variance comparison in gradient variance fraction calculation.
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)
        self.variance_threshold = variance_threshold

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: TensorStatistics model or a float.
        """

        gradients = [p.grad for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None]

        if not gradients:
            return 0.0  # Return 0.0 instead of None for no gradients

        # Calculate variance fraction
        return self._calculate_variance_fraction(gradients)

    def _calculate_variance_fraction(self, gradients: list[torch.Tensor]) -> float:
        """
        Calculate the fraction of parameters with variance above threshold.
        fraction = sqrt(variance) >= threshold

        :param gradients: List of gradient tensors
        :return: Fraction of parameters with high variance (0.0 to 1.0)
        """
        total_parameters = 0
        variance_parameters = 0

        for grad in gradients:
            parameter_count = grad.numel()
            total_parameters += parameter_count

            variance = grad.var().item()

            if math.sqrt(variance) >= self.variance_threshold:
                variance_parameters += parameter_count

        if total_parameters == 0:
            return 0.0

        return variance_parameters / total_parameters


class AverageParameterUpdateMagnitudeStatistics(Statistic):

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None or a float.
        """

        update_tensor = observation_utils.form_update_tensor(
            optimizer=optimizer, parameters=parameters, parameter_group=parameter_group
        )

        # when update tensor is none, return 0.0
        if update_tensor is None:
            return 0.0

        update_tensor = update_tensor.view(-1)
        update_tensor = update_tensor.abs()

        average_update_magnitude = update_tensor.mean().item()

        return average_update_magnitude


class MomentumGradientRatioStatistics(Statistic):

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """

        momentum = observation_utils.form_momentum_tensor(
            optimizer=optimizer, parameters=parameters, parameter_group=parameter_group
        )
        if momentum is None:
            return None

        gradients_list = [
            p.grad.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        # Handle empty gradients list
        if not gradients_list:
            return 0.0

        gradients = torch.cat(gradients_list).view(-1)

        # momentum_gradient_ratio r^t=\frac{\|g^t\|_2}{\|\nabla f(w^t)\|_2}
        gradients_norm = gradients.norm(p=2)
        momentum_norm = momentum.norm(p=2)

        if momentum_norm == 0:
            momentum_gradient_ratio = 0.0
        else:
            momentum_gradient_ratio = (gradients_norm / momentum_norm).item()

        return momentum_gradient_ratio


class LogOfNoiseScaleStatistics(Statistic):
    """
    Statistics for the log of noise scale in training.

    Tracks the log of noise scale B_{noise} using the formula:
    B_{noise} = tr(ΣH) / (G^T H G) ≈ (B/ε) * tr(HΣ) / tr(H^3 Σ)
    where:
    - H is the Hessian matrix
    - G is the gradient vector
    - Σ is the noise covariance matrix
    - B is the batch size
    - ε is the learning rate
    """

    @property
    def requires_gradient_graphs(self) -> bool:
        """
        :return: Whether the statistic requires gradient graphs to be retained.
        """

        return False

    @staticmethod
    def compute_hessian_diagonals(parameters: list[torch.Tensor]) -> torch.Tensor:
        """
        :param parameters: Parameters to compute the hessian diagonal matrices for.
        :return: Tensor containing the hessian diagonal matrices for all given parameters.
        """

        hessian_diagonals = []

        for parameter in parameters:
            if parameter.grad is not None:
                so_gradient = torch.autograd.grad(
                    outputs=parameter.grad.clone(),
                    inputs=parameter,
                    grad_outputs=torch.ones_like(parameter.grad, requires_grad=True),
                    only_inputs=True,
                    retain_graph=True,
                    create_graph=True,
                    allow_unused=True,
                )[0]

                if so_gradient is not None:
                    hessian_diagonals.append(so_gradient.view(-1))
                else:
                    hessian_diagonals.append(torch.zeros_like(parameter.view(-1)))

        return torch.cat(hessian_diagonals)

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.

        Computes the log of noise scale using the approximate formula:
        log(B_{noise}) ≈ log(B/ε) (move to observer) + log(tr(HΣ)) - log(tr(H^3 Σ))
        where:
        - H is the Hessian matrix
        - Σ is the noise covariance matrix
        - B is the batch size
        - ε is the learning rate

        """

        # Compute Hessian diagonals as in SecondOrderGradients Observation
        # hessian_diagonals = self.compute_hessian_diagonals(parameters)
        # use squared first order gradients as approximations
        fo_gradients = [
            p.grad.view(-1) for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]
        if not fo_gradients:
            return None

        hessian_diagonals = torch.cat(fo_gradients) ** 2

        if hessian_diagonals.numel() == 0:  # No gradients
            return None

        # For noise covariance matrix Σ, we'll use the identity matrix as an approximation
        # This is a common assumption when the exact noise structure is unknown
        noise_covariance = torch.ones_like(hessian_diagonals)

        # Compute tr(HΣ), add zero division tolerance to avoid log of zero when gradient is too small
        trace_hessian_noise_covariance = (
            torch.sum(hessian_diagonals * noise_covariance) + LHOPT_CONSTANTS["ZERO_DIVISION_TOLERANCE"]
        )

        log_trace_hessian_noise_covariance = torch.log(trace_hessian_noise_covariance).item()

        # Compute tr(H^3 Σ), add zero division tolerance to avoid log of zero when gradient is too small
        trace_hessian_cubed_noise_covariance = (
            torch.sum(hessian_diagonals**3 * noise_covariance) + LHOPT_CONSTANTS["ZERO_DIVISION_TOLERANCE"]
        )

        log_trace_hessian_cubed_noise_covariance = torch.log(trace_hessian_cubed_noise_covariance).item()

        # Compute final result: log(B_{noise}) ≈ log(tr(HΣ)) - log(tr(H^3 Σ))
        # Note: log(B/ε) term is handled in the observer layer
        log_noise_scale_without_log_b_over_epsilon = (
            log_trace_hessian_noise_covariance - log_trace_hessian_cubed_noise_covariance
        )

        return log_noise_scale_without_log_b_over_epsilon


class CosineSimilarityObserverOfGradientAndMomentumStatistics(Statistic):
    """
    Statistics for the cosine similarity of gradient and momentum.
    """

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """
        parameters_with_grads = [
            p for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        if not parameters_with_grads:
            return None

        gradients_list = [p.grad.view(-1) for p in parameters_with_grads]
        gradients = torch.cat(gradients_list).view(-1)

        momentum = observation_utils.form_momentum_tensor(
            optimizer=optimizer, parameters=parameters_with_grads, parameter_group=parameter_group
        )
        if momentum is None:
            return None

        gradients_2d = gradients.unsqueeze(0)
        momentum_2d = momentum.unsqueeze(0)

        cosine_similarity = F.cosine_similarity(gradients_2d, momentum_2d, dim=1).item()

        return cosine_similarity


class CosineSimilarityObserverOfGradientAndUpdateStatistics(Statistic):
    """
    Statistics for the cosine similarity of gradient and update.
    """

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """
        # Filter parameters that have gradients to ensure consistent tensor sizes
        parameters_with_grads = [
            p for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        if not parameters_with_grads:
            return None

        gradients_list = [p.grad.view(-1) for p in parameters_with_grads]
        gradients = torch.cat(gradients_list).view(-1)

        update_tensor = observation_utils.form_update_tensor(
            optimizer=optimizer, parameters=parameters_with_grads, parameter_group=parameter_group
        )

        if update_tensor is None:
            return None

        gradients_2d = gradients.unsqueeze(0)
        update_tensor_2d = update_tensor.unsqueeze(0)

        cosine_similarity = F.cosine_similarity(gradients_2d, update_tensor_2d, dim=1).item()

        return cosine_similarity


class CosineSimilarityOfGradientAndParameterStatistics(Statistic):
    """
    Statistics for the cosine similarity of gradient and parameter.
    """

    def _get_storage_format(self) -> StatisticStorageTypes:
        """
        :return: Storage format this observation stores data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _gather(
        self,
        *,
        optimizer: optim.Optimizer,
        model: nn.Module,
        parameters: list[torch.Tensor],
        parameter_group: dict[str, Any],
    ) -> torch.Tensor | TensorStatistics | float | None:
        """
        :param optimizer: Optimizer the given parameters and parameter group came from.
        :param model: Inner model to gather statistics from.
        :param parameters: List of parameters to gather statistics from.
        :param parameter_group: Parameter group the parameters originate from.
        :return: None, TensorStatistics model or a float.
        """
        # Filter parameters that have gradients to ensure consistent tensor sizes
        parameters_with_grads = [
            p for p in parameters if observation_utils.tensor_on_local_rank(p) and p.grad is not None
        ]

        if not parameters_with_grads:
            return None

        gradients_list = [p.grad.view(-1) for p in parameters_with_grads]
        gradients = torch.cat(gradients_list).view(-1)

        parameters_list = [p.view(-1) for p in parameters_with_grads]

        if not parameters_list:
            return None

        parameters_tensor = torch.cat(parameters_list).view(-1)

        gradients_2d = gradients.unsqueeze(0)
        parameters_tensor_2d = parameters_tensor.unsqueeze(0)

        cosine_similarity = F.cosine_similarity(gradients_2d, parameters_tensor_2d, dim=1).item()

        return cosine_similarity



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================
import math
from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import LHOPTCheckpointBaseObserver
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TrainingProgress(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.training_progress

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class EpochsCompleted(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.epochs_completed

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ProgressAtEachCheckpoint(LHOPTCheckpointBaseObserver):
    """
    This is a global observer from the paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns a single float value that is the training progress at the current checkpoint.
    The observation tracks training progress and returns the progress value only when a checkpoint is reached.
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float:
        """
        Returns training progress at each checkpoint interval.

        Before checkpoint: returns progress towards next checkpoint (0 to 1)
        At checkpoint: returns the actual training progress value
        """
        current_progress = observation_inputs.training_progress

        # Cold start: If the last progress is not set, set it to the first progress record
        self._cold_start(current_progress)

        self._update_history(current_progress)

        # Check if we should create a checkpoint
        if self._should_create_checkpoint():
            # Return the progress at this checkpoint
            self.last_value = current_progress
            return current_progress
        else:
            return self.last_value


class StagnationObserver(GlobalObserver):

    def __init__(self, stagnation_threshold: float = 0.0, log_transform: bool = True, **kwargs):
        """
        :param stagnation_threshold: The loss improvement threshold. Default is 0.0. This is the amount the validation
        loss must improve by over the best validation loss to reset the stagnation counter.
        We assume that improvement means loss decreases regardless of the sign of the loss, which means the best loss is
        negative infinity.
        :param log_transform: Whether to log transform the stagnation counter. Default is True.
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(**kwargs)
        self.best_validation_loss: float | None = None
        self.stagnation_counter: int = 0
        self.stagnation_threshold = stagnation_threshold
        self.log_transform = log_transform

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """
        # Validate input values, if loss is invalid, increment stagnation counter
        if math.isnan(observation_inputs.validation_loss) or math.isinf(observation_inputs.validation_loss):
            self.stagnation_counter += 1
        else:
            if self.best_validation_loss is None:
                self.best_validation_loss = observation_inputs.validation_loss
                self.stagnation_counter = 0
            else:
                improvement = self.best_validation_loss - observation_inputs.validation_loss
                if improvement > self.stagnation_threshold:
                    self.best_validation_loss = observation_inputs.validation_loss
                    self.stagnation_counter = 0
                else:
                    self.stagnation_counter += 1
        return self.stagnation_counter if not self.log_transform else math.log(self.stagnation_counter + 1)

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import copy
from typing import Any

import gymnasium as gym
import numpy as np

from libinephany.observations.observers.observer_containers import GlobalObserverContainer, LocalObserverContainer
from libinephany.observations.post_processors import postprocessors
from libinephany.observations.post_processors.postprocessors import ObservationPostProcessor
from libinephany.pydantic_models.configs.observer_config import AgentObserverConfig, ObserverConfig
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils.standardizers import Standardizer
from libinephany.utils.typing import ObservationInformation

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================
AGENT_OBSERVATION = "agent_observation"
AGENT_MASK = "agent_mask"

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class ObserverPipeline:

    def __init__(
        self,
        observer_config: ObserverConfig,
        agent_config: AgentObserverConfig,
        standardizer: Standardizer | None,
        agent_id_to_modules: dict[str, str | None],
    ):
        """
        :param observer_config: ObserverConfig that contains various parameters applicable to all agent observers as
        well as the observer configs for specific agents.
        :param agent_config: AgentObserverConfig storing configuration of the agent's actions and observation space for
        this type of agent.
        :param standardizer: None or the standardizer to apply to the returned observations.
        :param agent_id_to_modules: Dictionary mapping agent IDs to None or the name of the parameter group that agent
        is modulating.
        """

        self.observer_config = observer_config
        self.agent_config = agent_config
        self.standardizer = standardizer
        self.agent_id_to_modules = agent_id_to_modules

        self.global_observers = GlobalObserverContainer(
            global_config=observer_config,
            agent_config=agent_config,
            standardizer=standardizer,
        )

        self.local_observers: dict[str, LocalObserverContainer] = self._build_local_observers(
            agent_id_to_modules=agent_id_to_modules,
        )

        self.post_processors: list[ObservationPostProcessor] = self._build_post_processors()

        self.clipping_threshold = self.observer_config.observation_clipping_threshold
        self.invalid_threshold = self.observer_config.invalid_observation_threshold

    @property
    def observation_size(self) -> int:
        """
        :return: Total size of the observation vectors produced by this pipeline.
        """

        globals_size = self.global_observers.total_observer_size
        locals_size = list(self.local_observers.values())[0].total_observer_size

        total_size = globals_size + locals_size

        if self.agent_config.prepend_invalid_indicator:
            total_size += 1

        return total_size

    def _build_local_observers(
        self,
        agent_id_to_modules: dict[str, str | None],
    ) -> dict[str, LocalObserverContainer]:
        """
        :param agent_id_to_modules: Dictionary mapping agent IDs to None or the name of the parameter group that agent
        is modulating.
        :return: Dictionary mapping agent IDs to LocalObserverContainers for each agent.
        """

        observers = {}

        for agent_id, parameter_group_name in agent_id_to_modules.items():
            observers[agent_id] = LocalObserverContainer(
                global_config=self.observer_config,
                agent_config=self.agent_config,
                standardizer=self.standardizer,
                agent_id=agent_id,
                parameter_group_name=parameter_group_name,
            )

        return observers

    def _build_post_processors(self) -> list[ObservationPostProcessor]:
        """
        :return: List of post-processors to apply to the global and local observations.
        """

        if self.agent_config.postprocessors is None:
            return []

        post_processors = []

        for observer_name, post_processor_kwargs in self.agent_config.postprocessors.items():
            try:
                post_processor_type: type[ObservationPostProcessor] = getattr(postprocessors, observer_name)

            except AttributeError as e:
                raise AttributeError(f"The class {observer_name} does not exist within {postprocessors}!") from e

            if post_processor_kwargs is None:
                post_processor_kwargs = {}

            post_processor = post_processor_type(
                observer_config=self.observer_config,
                **post_processor_kwargs,
            )

            post_processors.append(post_processor)

        return post_processors

    @staticmethod
    def merge_globals_to_locals(
        global_obs: list[float | int],
        local_obs: dict[str, list[float | int]],
    ) -> dict[str, list[float | int]]:
        """
        :param global_obs: Global observations to post-process.
        :param local_obs: Dictionary mapping agent IDs to local observations of that agent.
        :return: Tuple of clipped global and local observations.
        :return: Dictionary mapping agent ID to that agent's completed observation vector.
        """

        return {agent_id: global_obs + agent_obs for agent_id, agent_obs in local_obs.items()}

    def get_required_trackers(self) -> list[dict[str, dict[str, Any] | None]]:
        """
        :return: List of trackers required by each of the stored observers.
        """

        required_trackers = self.global_observers.get_required_trackers()

        for agent_container in self.local_observers.values():
            required_trackers += agent_container.get_required_trackers()

        return required_trackers

    def clip_observation_vector(self, observation_vector: list[float | int]) -> tuple[bool, list[float | int]]:
        """
        :param observation_vector: Observations to clip.
        :return: Tuple indicating whether an observation was clipped and a list of the post-processed observations.
        """

        invalid_encountered = False
        post_processed = []

        for observation in observation_vector:
            if abs(observation) >= self.clipping_threshold:
                if abs(observation) >= self.invalid_threshold:
                    invalid_encountered = True

                observation = -self.clipping_threshold if observation < 0 else self.clipping_threshold

            elif np.isnan(observation):
                observation = self.observer_config.invalid_observation_replacement_value
                invalid_encountered = True

            post_processed.append(observation)

        return invalid_encountered, post_processed

    def clip_observations(
        self,
        globals_to_clip: list[float | int],
        locals_to_clip: dict[str, list[float | int]],
    ) -> tuple[list[float | int], dict[str, list[float | int]], bool]:
        """
        :param globals_to_clip: Global observations to post-process.
        :param locals_to_clip: Dictionary mapping agent IDs to local observations of that agent.
        :return: Tuple of clipped global and local observations and whether any observation clipping occurred.
        """

        invalid_encountered, globals_to_clip = self.clip_observation_vector(observation_vector=globals_to_clip)
        post_processed_locals = {}

        for agent_id, agent_observations in locals_to_clip.items():
            agent_invalid_encountered, post_processed = self.clip_observation_vector(
                observation_vector=agent_observations
            )

            post_processed_locals[agent_id] = post_processed
            invalid_encountered = invalid_encountered if invalid_encountered else agent_invalid_encountered

        if self.agent_config.prepend_invalid_indicator:
            globals_to_clip = [int(invalid_encountered)] + globals_to_clip

        return globals_to_clip, post_processed_locals, invalid_encountered

    def observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        actions_taken: dict[str, float | int | None],
        return_dict: bool = False,
    ) -> tuple[
        dict[str, list[float | int]], bool, dict[str, dict[str, list[float | int] | dict[str, float | int]]] | None
    ]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param actions_taken: Dictionary mapping agent IDs to actions taken by that agent.
        :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
        :return: Tuple of:
         - A dictionary mapping agent ID to that agent's completed observation vector,
         - A boolean indicating whether an observation clip occurred,
         - A dictionary mapping agent ID to a dictionary mapping observer name to that observer's observation vector.
        """

        global_obs, global_obs_dict = self.global_observers.observe(
            observation_inputs=observation_inputs,
            hyperparameter_states=hyperparameter_states,
            tracked_statistics=tracked_statistics,
            action_taken=None,
            return_dict=return_dict,
        )

        local_obs: dict[str, list[float | int]] = {}
        obs_as_dict: dict[str, dict[str, list[float | int] | dict[str, float | int]]] = {}

        for agent_id, agent_observers in self.local_observers.items():
            local_obs[agent_id], local_obs_dict = agent_observers.observe(
                observation_inputs=observation_inputs,
                hyperparameter_states=hyperparameter_states,
                tracked_statistics=tracked_statistics,
                action_taken=actions_taken[agent_id] if agent_id in actions_taken else None,
                return_dict=return_dict,
            )

            if return_dict:
                obs_as_dict[agent_id] = copy.deepcopy(global_obs_dict)
                obs_as_dict[agent_id].update(local_obs_dict)

        for post_processor in self.post_processors:
            global_obs, local_obs = post_processor.postprocess(
                global_observations=global_obs, local_observations=local_obs
            )

        global_obs, local_obs, obs_clipped = self.clip_observations(
            globals_to_clip=global_obs, locals_to_clip=local_obs
        )
        merged_obs = self.merge_globals_to_locals(global_obs=global_obs, local_obs=local_obs)

        return merged_obs, obs_clipped, obs_as_dict if return_dict else None

    def inform(self) -> tuple[ObservationInformation, dict[str, ObservationInformation]]:
        """
        :return: Dictionary of observation info to add to the agent info.
        """

        global_information = self.global_observers.inform()

        agent_information = {agent_id: observer.inform() for agent_id, observer in self.local_observers.items()}

        return global_information, agent_information

    def get_observation_spaces(self) -> dict[str, dict[str, gym.spaces.Box]]:
        """
        :return: Dictionary mapping agent IDs to their observation spaces.
        """

        return {
            agent_id: {
                AGENT_OBSERVATION: gym.spaces.Box(
                    low=-np.inf, high=np.inf, shape=(self.observation_size,), dtype=np.float32
                ),
                AGENT_MASK: gym.spaces.Box(low=0, high=1, shape=(), dtype=np.float32),
            }
            for agent_id in self.local_observers
        }

    def reset(self) -> None:
        """
        Resets all global and local observers.
        """

        self.global_observers.reset()

        for local_observers in self.local_observers.values():
            local_observers.reset()

    def train(self) -> None:
        """
        Sets all observer containers into training mode.
        """

        self.global_observers.train()

        for local_observers in self.local_observers.values():
            local_observers.train()

    def infer(self) -> None:
        """
        Sets all observer containers into inference mode.
        """

        self.global_observers.infer()

        for local_observers in self.local_observers.values():
            local_observers.infer()



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any

import gymnasium as gym

from libinephany.observations.observer_pipeline import ObserverPipeline
from libinephany.pydantic_models.configs.observer_config import ObserverConfig
from libinephany.pydantic_models.schemas.agent_info import AgentInfo
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs, Observations
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils import standardizers

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class ObserverPipelineCoordinator:

    def __init__(
        self,
        observer_config: ObserverConfig,
    ):
        """
        :param observer_config: ObserverConfig that contains various parameters applicable to all agent observers as
        well as the observer configs for specific agents.
        """

        self.observer_config = observer_config
        self.standardizer = standardizers.get_standardizer(
            standardizer_name=self.observer_config.standardizer,
            standardizer_arguments=self.observer_config.standardizer_arguments,
        )

        self.pipelines: list[ObserverPipeline] = self._build_pipelines()

    @staticmethod
    def _merge_pipeline_requirements(
        pipeline: ObserverPipeline, required_trackers: dict[str, dict[str, Any] | None]
    ) -> dict[str, dict[str, Any] | None]:
        """
        :param pipeline: ObserverPipeline to compile the tracker requirements of its observers.
        :param required_trackers: Running dictionary of merged requirements.
        :return: Merged tracker requirements.
        """

        pipeline_requirements = pipeline.get_required_trackers()

        for observer_requirement in pipeline_requirements:
            for tracker_name, tracker_kwargs in observer_requirement.items():
                if tracker_name in required_trackers and tracker_kwargs != required_trackers[tracker_name]:
                    raise ValueError(
                        f"Conflict for tracker '{tracker_name}': "
                        f"Existing value: {required_trackers[tracker_name]}, "
                        f"New value: {tracker_kwargs}"
                    )

                else:
                    required_trackers[tracker_name] = tracker_kwargs

        return required_trackers

    def _build_pipelines(self) -> list[ObserverPipeline]:
        """
        :return: List of pipelines for each agent type.
        """

        return [
            ObserverPipeline(
                observer_config=self.observer_config,
                agent_config=agent_config,
                standardizer=self.standardizer,
                agent_id_to_modules=agent_id_to_modules,
            )
            for (agent_config, agent_id_to_modules) in self.observer_config.agents_to_modules_by_type
        ]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping tracker names to None or the kwargs for that tracker.
        """

        required_trackers: dict[str, dict[str, Any] | None] = {}

        for pipeline in self.pipelines:
            required_trackers = self._merge_pipeline_requirements(
                pipeline=pipeline, required_trackers=required_trackers
            )

        return required_trackers

    def get_observation_space(self) -> gym.spaces.Dict:
        """
        :return: Gymnasium dictionary space mapping Agent IDs to observation spaces.
        """

        spaces: dict[str, gym.spaces.Space] = {}

        for pipeline in self.pipelines:
            observations_spaces = pipeline.get_observation_spaces()

            spaces.update(**observations_spaces)  # type: ignore[call-overload]

        if not spaces:
            raise ValueError("No observation spaces have been created!")

        spaces = {k: gym.spaces.Dict(v) for k, v in spaces.items()}  # type: ignore[arg-type]

        return gym.spaces.Dict(spaces)

    def observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        actions_taken: dict[str, float | int | None],
        return_dict: bool = False,
    ) -> Observations:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param actions_taken: Dictionary mapping agent IDs to actions taken by that agent.
        :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
        :return: Observations pydantic model.
        """

        clipped_observations = False
        observations = {}
        observations_as_dict: dict[str, dict[str, list[float | int] | dict[str, float | int]]] = {}

        for pipeline in self.pipelines:
            pipeline_observations, pipeline_clipped_observations, pipeline_observations_dict = pipeline.observe(
                observation_inputs=observation_inputs,
                hyperparameter_states=hyperparameter_states,
                tracked_statistics=tracked_statistics,
                actions_taken=actions_taken,
                return_dict=return_dict,
            )

            clipped_observations = clipped_observations if clipped_observations else pipeline_clipped_observations

            for agent_id, agent_obs in pipeline_observations.items():
                if agent_id in observations:
                    raise KeyError(
                        f"Attempted to add observations for agent {agent_id} to observations dictionary but they are "
                        f"already present! This should not happen under any circumstances."
                    )

                observations[agent_id] = agent_obs

                if return_dict:
                    observations_as_dict[agent_id] = pipeline_observations_dict[agent_id]

        return Observations(
            observation_inputs=observation_inputs,
            agent_observations=observations,
            hit_invalid_value=clipped_observations,
            observations_as_dict=observations_as_dict if return_dict else None,
        )

    def inform(self) -> AgentInfo:
        """
        :return: AgentInfo model storing information about various global and local observations.
        """

        information = AgentInfo()

        for pipeline in self.pipelines:
            pipeline_global_information, pipeline_agent_information = pipeline.inform()

            information.update_global_information(new_information=pipeline_global_information)
            information.add_agent_information(agent_information=pipeline_agent_information)

        return information

    def reset(self) -> None:
        """
        Resets all observer pipelines.
        """

        for pipeline in self.pipelines:
            pipeline.reset()

    def train(self) -> None:
        """
        Sets all observer pipelines into training mode.
        """

        for pipeline in self.pipelines:
            pipeline.train()

    def infer(self) -> None:
        """
        Sets all observer pipelines into inference mode.
        """

        for pipeline in self.pipelines:
            pipeline.infer()



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from enum import Enum
from itertools import chain
from typing import Any, Callable

import numpy as np
import pandas as pd
import torch
import torch.optim as optim
from loguru import logger
from scipy.stats import norm

from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.utils import optim_utils

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

EXP_AVERAGE = "exp_avg"
MOMENTUM_BUFFER = "momentum_buffer"
MIN_DECAY_FACTOR = 1e-10

MIN_TOTAL_WEIGHT = 1e-15  # Minimum total weight threshold for numerical stability

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class StatisticsCallStage(Enum):

    ON_BATCH_END = "on_batch_end"
    ON_OPTIMIZER_STEP = "on_optimizer_step"
    ON_TRAIN_END = "on_train_end"

    FORWARD_HOOK = "forward_hook"


class StatisticStorageTypes(Enum):

    TENSOR_STATISTICS = TensorStatistics.__name__
    FLOAT = float.__name__
    VECTOR = "vector"


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def get_exponential_weighted_average(
    values: list[int | float],
    invalid_value_threshold: float = 1e10,
    tracker_name: str = "Unknown",
    alpha: float = 0.1,
) -> float:
    """
    :param values: List of values to average via EWA.
    :param invalid_value_threshold: Threshold for invalid observations, default is 1e10.
    :param tracker_name: Name of the tracker for error reporting, default is "Unknown".
    :param alpha: Alpha to use whe computing the exponentially weighted average.
    :return: EWA of the given values.
    """

    # Logging statistic tracker name when input is invalid.

    if len(values) == 0:
        raise ValueError(
            f"Statistic Tracker: {tracker_name} gathered data with empty list! It likely means a bug is triggered in the"
            f" code related to {tracker_name} or the inner task. Please check up the code!"
        )

    if any(not isinstance(value, (float, int)) for value in values):
        # Should never happen, but just in case
        raise ValueError(
            f"Statistic Tracker: {tracker_name} gathered data with invalid values (not int or float)! It likely means a"
            f" bug is triggered in the code related to {tracker_name} or the inner task. Please check up the code!"
        )

    if any(abs(value) > invalid_value_threshold for value in values):
        # check for large values (including negative values)
        logger.warning(
            f"Statistic Tracker: {tracker_name} gathered data with values out of range (invalid_value_threshold: "
            f"{invalid_value_threshold})! May cause episode termination if StopEpisodeFromInvalidObservations is used."
        )

    if any(np.isnan(value) for value in values):
        logger.warning(
            f"Statistic Tracker: {tracker_name} gathered data with NaN values! May cause episode termination if "
            f"StopEpisodeFromInvalidObservations is used."
        )

    exp_weighted_average = pd.Series(values).ewm(alpha=alpha).mean().iloc[-1]
    assert isinstance(exp_weighted_average, float)
    return exp_weighted_average


def apply_averaging_function_to_tensor_statistics(
    tensor_statistics: list[TensorStatistics],
    averaging_function: Callable[[list[float], float, str], float],
    invalid_value_threshold: float = 1e10,
    tracker_name: str = "Unknown",
) -> TensorStatistics:
    """
    :param tensor_statistics: List of statistics models to average over.
    :param averaging_function: Function to average the values with.
    :return: TensorStatistics containing the average over all given tensor statistics.
    """

    fields = TensorStatistics.model_fields.keys()
    averaged_metrics = {
        field: averaging_function(
            [getattr(statistics, field) for statistics in tensor_statistics],
            invalid_value_threshold,
            tracker_name,
        )
        for field in fields
    }

    return TensorStatistics(**averaged_metrics)


def apply_averaging_function_to_dictionary_of_tensor_statistics(
    data: dict[str, list[TensorStatistics]],
    averaging_function: Callable[[list[float], float, str], float],
    invalid_value_threshold: float = 1e10,
    tracker_name: str = "Unknown",
) -> dict[str, TensorStatistics]:
    """
    :param data: Dictionary mapping parameter group names to list of TensorStatistics from that parameter group.
    :param averaging_function: Function to average the values with.
    :param invalid_value_threshold: Threshold for invalid observations.
    :param tracker_name: Name of the tracker for error reporting.
    :return: Dictionary mapping parameter group names to TensorStatistics averaged over all statistics in the given
    TensorStatistics models.
    """

    return {
        group: apply_averaging_function_to_tensor_statistics(
            tensor_statistics=metrics,
            averaging_function=averaging_function,
            invalid_value_threshold=invalid_value_threshold,
            tracker_name=tracker_name,
        )
        for group, metrics in data.items()
    }


def apply_averaging_function_to_dictionary_of_metric_lists(
    data: dict[str, list[float]],
    averaging_function: Callable[[list[float], float, str], float],
    invalid_value_threshold: float = 1e10,
    tracker_name: str = "Unknown",
) -> dict[str, float]:
    """
    :param data: Dictionary mapping parameter group names to list of metrics from that parameter group.
    :param averaging_function: Function to average the values with.
    :param invalid_value_threshold: Threshold for invalid observations.
    :param tracker_name: Name of the tracker for error reporting.
    :return: Dictionary mapping parameter group names to averages over all metrics from each parameter group.
    """

    return {
        group: averaging_function(metrics, invalid_value_threshold, tracker_name) for group, metrics in data.items()
    }


def average_tensor_statistics(tensor_statistics: list[TensorStatistics]) -> TensorStatistics:
    """
    :param tensor_statistics: List of TensorStatistics models to average into one model.
    :return: Averages over all given tensor statistics models.
    """

    averaged = {
        field: sum([getattr(statistics_model, field) for statistics_model in tensor_statistics])
        for field in TensorStatistics.model_fields.keys()
    }
    averaged = {field: total / len(tensor_statistics) for field, total in averaged.items()}

    return TensorStatistics(**averaged)


def create_one_hot_observation(vector_length: int, one_hot_index: int | None) -> list[int | float]:
    """
    :param vector_length: Length of the one-hot vector.
    :param one_hot_index: Index of the vector whose element should be set to 1.0, leaving all others as 0.0.
    :return: Constructed one-hot vector in a list.
    """

    if one_hot_index is not None and one_hot_index < 0:
        raise ValueError("One hot indices must be greater than 0.")

    one_hot = np.zeros(vector_length, dtype=np.int8)

    if one_hot_index is not None:
        one_hot[one_hot_index] = 1

    as_list = one_hot.tolist()

    assert isinstance(as_list, list), "One-hot vector must be a list."

    return as_list


def create_one_hot_depth_encoding(agent_controlled_modules: list[str], parameter_group_name: str) -> list[int | float]:
    """
    :param agent_controlled_modules: Ordered list of parameter group names in the inner model.
    :param parameter_group_name: Name of the parameter group to create a depth one-hot vector for.
    :return: Constructed one-hot depth encoding in a list.

    :note: GANNO encodes depths to one-hot vectors of length 3 regardless of the size of the model.
    """

    module_index = agent_controlled_modules.index(parameter_group_name)
    number_of_modules = len(agent_controlled_modules)

    one_hot_index = min(2, (module_index * 3) // number_of_modules)

    return create_one_hot_observation(vector_length=3, one_hot_index=one_hot_index)


def check_if_tensor_is_xpu(tensor: torch.Tensor | None) -> bool:
    """
    :param tensor: Tensor to check whether it is an XPU tensor.
    :return: Whether the tensor is an XPU tensor.
    """

    try:
        xpu_available = torch.xpu.is_available()
        return xpu_available and tensor.is_xpu

    except AttributeError:
        return False


def tensor_on_local_rank(tensor: torch.Tensor | None) -> bool:
    """
    :param tensor: Tensor to check whether it is owned by the local rank, partially or entirely.
    :return: Whether the tensor is owned by the local rank.
    """

    valid_tensor = tensor is not None and tensor.numel() > 0
    device_index = tensor.device.index if valid_tensor else None

    is_xpu = check_if_tensor_is_xpu(tensor)
    cuda_available = torch.cuda.is_available()

    if valid_tensor and is_xpu:
        current_device = torch.xpu.current_device()
        return device_index == current_device

    elif valid_tensor and cuda_available and tensor.is_cuda:
        current_device = torch.cuda.current_device()
        return device_index == current_device

    return valid_tensor


def form_update_tensor(
    optimizer: optim.Optimizer, parameters: list[torch.Tensor], parameter_group: dict[str, Any]
) -> None | torch.Tensor:
    """
    :param optimizer: Optimizer to form the update tensor from.
    :param parameters: Parameters to create the update tensor from.
    :param parameter_group: Parameter group within the optimizer the given parameters came from.
    :return: None or the formed update tensor.
    """

    if type(optimizer) in optim_utils.ADAM_OPTIMISERS:
        updates_list = [optimizer.state[p][EXP_AVERAGE].view(-1) for p in parameters if tensor_on_local_rank(p)]
        return torch.cat(updates_list) if updates_list else None

    elif type(optimizer) in optim_utils.SGD_OPTIMISERS:
        return optim_utils.compute_sgd_optimizer_update_stats(
            optimizer=optimizer, parameter_group=parameter_group, parameters=parameters
        )

    else:
        raise NotImplementedError(f"Optimizer {type(optimizer).__name__} is not supported!")


def form_momentum_tensor(
    optimizer: optim.Optimizer, parameters: list[torch.Tensor], parameter_group: dict[str, Any]
) -> None | torch.Tensor:
    """
    :param optimizer: Optimizer to form the momentum tensor from.
    :param parameters: Parameters to create the momentum tensor from.
    :param parameter_group: Parameter group within the optimizer the given parameters came from.
    """
    if type(optimizer) in optim_utils.ADAM_OPTIMISERS:
        momentum_list = [optimizer.state[p][EXP_AVERAGE].view(-1) for p in parameters if tensor_on_local_rank(p)]
        return torch.cat(momentum_list) if momentum_list else None
    elif type(optimizer) in optim_utils.SGD_OPTIMISERS:
        momentum_list = [optimizer.state[p][MOMENTUM_BUFFER].view(-1) for p in parameters if tensor_on_local_rank(p)]
        return torch.cat(momentum_list) if momentum_list else None
    else:
        raise NotImplementedError(f"Optimizer {type(optimizer).__name__} is not supported!")


def null_standardizer(value_to_standardize: float, **kwargs) -> float:
    """
    :param value_to_standardize: Value to mock the standardization of.
    :return: Given value to standardize.
    """

    return value_to_standardize


def create_sinusoidal_depth_encoding(
    agent_controlled_modules: list[str], parameter_group_name: str, dimensionality: int
) -> list[int | float]:
    """
    :param agent_controlled_modules: Ordered list of parameter group names in the inner model.
    :param parameter_group_name: Name of the parameter group to create a depth encoding for.
    :param dimensionality: Length of the depth vector.
    :return: Sinusoidal depth encoding.
    """

    assert dimensionality % 2 == 0, "Dimensionality of a sinusoidal depth encoding must be even."

    depth = agent_controlled_modules.index(parameter_group_name)

    positions = np.arange(dimensionality // 2)
    frequencies = 1 / (10000 ** (2 * positions / dimensionality))

    encoding = np.zeros(dimensionality)
    encoding[0::2] = np.sin(depth * frequencies)
    encoding[1::2] = np.cos(depth * frequencies)

    vector = encoding.tolist()

    return vector


def concatenate_lists(lists: list[list[Any]]) -> list[Any]:
    """
    :param lists: Lists to concatenate.
    :return: Concatenated lists.
    """

    return list(chain(*lists))


def compute_cdf_weighted_mean_and_std(
    time_series: list[tuple[float, float]], decay_factor: float
) -> tuple[float, float]:
    """
    Compute the CDF-weighted standard deviation using the same exponential decay weights
    as the mean calculation, with numerical integration.

    :param time_series: List of (time, value) pairs
    :param decay_factor: Decay factor b in the exponential weight formula b in [1.25, 2.5, 5, 10, 20]
    :return: Tuple of (weighted mean, weighted standard deviation)
    """

    if len(time_series) == 0:
        return 0.0, 0.0

    if len(time_series) == 1:
        return time_series[0][1], 0.0

    sorted_series = sorted(time_series, key=lambda x: x[0])

    # Handle the special case when decay_factor = 1.0
    if abs(decay_factor - 1.0) < MIN_DECAY_FACTOR:
        # When decay_factor = 1.0, w(t) = 1 for all t
        # So the result is just the arithmetic mean
        values = [v for _, v in sorted_series]
        mean = float(np.mean(values))
        std = float(np.std(values))
        return mean, std

    log_decay_factor = math.log(decay_factor)

    total_weighted_value = 0.0  # ∫ w(t) y(t) dt - total weighted value
    total_weighted_squared = 0.0  # ∫ w(t) y(t)² dt - total weighted squared value

    for time_series_index in range(len(sorted_series) - 1):
        start_time_point = sorted_series[time_series_index][0]
        end_time_point = sorted_series[time_series_index + 1][0]
        start_value = sorted_series[time_series_index][1]
        end_value = sorted_series[time_series_index + 1][1]

        time_interval = end_time_point - start_time_point
        assert time_interval > 0, "Time interval must be positive"

        interval_value = _weighted_interval_expectation(
            start_time_point=start_time_point,
            start_value=start_value,
            end_time_point=end_time_point,
            end_value=end_value,
            log_decay_factor=log_decay_factor,
        )
        interval_squared_value = _weighted_interval_expectation(
            start_time_point=start_time_point,
            start_value=start_value**2,
            end_time_point=end_time_point,
            end_value=end_value**2,
            log_decay_factor=log_decay_factor,
        )

        total_weighted_value += interval_value
        total_weighted_squared += interval_squared_value

    total_weight = (1 / log_decay_factor) * (
        math.exp(log_decay_factor * sorted_series[-1][0]) - math.exp(log_decay_factor * sorted_series[0][0])
    )
    # Check if total weight is too small (numerical stability)
    if total_weight < MIN_TOTAL_WEIGHT:
        values = [v for _, v in sorted_series]
        mean = float(np.mean(values))
        std = float(np.std(values))
        return mean, std

    # Calculate weighted mean: μ = ∫ w(t) y(t) dt / ∫ w(t) dt
    # This gives us the expected value under the weight distribution
    weighted_mean = float(total_weighted_value / total_weight)

    # Calculate weighted variance: Var = ∫ w(t) y(t)² dt / ∫ w(t) dt - μ²
    # This follows from the definition: Var(X) = E[X²] - (E[X])²
    # where E[X] = ∫ w(t) y(t) dt / ∫ w(t) dt and E[X²] = ∫ w(t) y(t)² dt / ∫ w(t) dt
    weighted_variance = float(total_weighted_squared / total_weight - weighted_mean**2)

    # Calculate weighted standard deviation: σ = √Var
    # This is the square root of the variance, representing the spread of values
    weighted_std = float(math.sqrt(max(0, weighted_variance)))

    return weighted_mean, weighted_std


def _weighted_interval_expectation(
    start_time_point: float,
    start_value: float,
    end_time_point: float,
    end_value: float,
    log_decay_factor: float,
) -> float:
    """
    Computes the weighted interval expectation from Appendix E of the LHOPT paper. NB: that paper has a typo in the
    integral, y_n and y_{n+1} should be swapped.

    :param start_time_point: the start time value of the interval.
    :param start_value: the value at start_time_point.
    :param end_time_point: the end time value of the interval.
    :param end_value: the value at end_time_point.
    :param log_decay_factor: the logarithm of the decay factor used to weight the expectation.
    :return: the exponentially-weighted expectation of the linear interpolation between the start and end points.
    """

    interval_gradient = (end_value - start_value) / (end_time_point - start_time_point)
    start_exp_time = math.exp(log_decay_factor * start_time_point)
    end_exp_time = math.exp(log_decay_factor * end_time_point)
    return (1 / log_decay_factor) * (end_value * end_exp_time - start_value * start_exp_time) - (
        1 / log_decay_factor**2
    ) * interval_gradient * (end_exp_time - start_exp_time)


def compute_cdf_feature(
    current_value: float,
    time_series: list[tuple[float, float]],
    decay_factor: float,
    current_time: float,
    time_window: int,
) -> float:
    """

    This function computes a CDF feature that represents the cumulative probability
    of the current value given the historical distribution, weighted by time decay.
    Uses scipy.stats.norm.cdf with loc (mean) and scale (std) computed from CDF utilities.

    The mean and std formula from the OpenAI paper:
    https://arxiv.org/pdf/2305.18290.pdf


    :param current_value: Current value to compute CDF feature for
    :param time_series: List of (time, value) pairs for CDF calculation. time_series will be updated in-place each time this function is called.
    :param decay_factor: Decay factor for CDF calculation (0 < factor < 1)
    :param current_time: Current time step
    :param time_window: Maximum number of time steps to keep in time series
    :return: CDF feature value (cumulative probability from normal distribution)
    """
    # Add current observation to time series
    time_series.append((current_time, current_value))

    # Keep only the last time_window observations
    if len(time_series) > time_window:
        time_series[:] = time_series[-time_window:]

    # If we don't have enough data, return 0.0
    if len(time_series) < 2:
        return 0.0

    # Compute CDF-weighted mean (loc) and standard deviation (scale)
    cdf_mean, cdf_std = compute_cdf_weighted_mean_and_std(time_series, decay_factor)

    # Compute CDF feature using scipy.stats.norm.cdf
    if cdf_std > 0:
        # Use norm.cdf with loc=cdf_mean and scale=cdf_std
        cdf_feature = norm.cdf(current_value, loc=cdf_mean, scale=cdf_std)
        return cdf_feature
    else:
        # If the standard deviation is 0, return 0.0
        return 0.0
